#include <stdio.h>

#include <glib.h>
#include <glib/gi18n.h>

#include <libnotify/notify.h>

#include "bognor-marshal.h"
#include "bognor-queue.h"
#include "bognor-queue-glue.h"
#include "bognor-source-manager.h"

enum {
    PROP_0,
    PROP_NAME,
    PROP_SOURCE,
};

enum {
    ADDED,
    REMOVED,
    INDEX_CHANGED,
    ITEM_CHANGED,
    POSITION_CHANGED,
    STATE_CHANGED,
    LAST_SIGNAL,
};

struct _BognorQueuePrivate {
    char *name; /* Name of the queue */
    GVolumeMonitor *volume_monitor; /* This probably doesn't need to be here
                                       and should be in a subclass or in the
                                       QueueMonitor */

    GQueue *play_queue; /* contains BognorQueueItem */
    GList *current_position; /* Owned by @play_queue */
    int current_index;

    int queue_duration; /* In seconds */

    BognorSourceManager *source_manager;

    BognorQueueState state;
    BognorQueueStatus status;
    BognorQueueMode mode;
};

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), BOGNOR_TYPE_QUEUE, BognorQueuePrivate))
G_DEFINE_TYPE (BognorQueue, bognor_queue, G_TYPE_OBJECT);
static guint32 signals[LAST_SIGNAL] = {0,};

GQuark
bognor_queue_error_quark (void)
{
    static GQuark quark = 0;

    if (G_UNLIKELY (quark == 0)) {
        quark = g_quark_from_static_string ("bognor-queue-error-quark");
    }

    return quark;
}

static void
bognor_queue_finalize (GObject *object)
{
    BognorQueue *self = (BognorQueue *) object;

    G_OBJECT_CLASS (bognor_queue_parent_class)->finalize (object);
}

static void
bognor_queue_dispose (GObject *object)
{
    BognorQueue *self = (BognorQueue *) object;
    BognorQueuePrivate *priv = self->priv;

    if (priv->source_manager) {
        g_object_unref (priv->source_manager);
        priv->source_manager = NULL;
    }

    G_OBJECT_CLASS (bognor_queue_parent_class)->dispose (object);
}

static void
bognor_queue_set_property (GObject      *object,
                           guint         prop_id,
                           const GValue *value,
                           GParamSpec   *pspec)
{
    BognorQueue *self = (BognorQueue *) object;
    BognorQueuePrivate *priv = self->priv;

    switch (prop_id) {
    case PROP_NAME:
        priv->name = g_value_dup_string (value);
        break;

    case PROP_SOURCE:
        priv->source_manager = g_value_get_object (value);
        break;

    default:
        break;
    }
}

static void
bognor_queue_get_property (GObject    *object,
                           guint       prop_id,
                           GValue     *value,
                           GParamSpec *pspec)
{
    BognorQueue *self = (BognorQueue *) object;

    switch (prop_id) {

    default:
        break;
    }
}

static void
bognor_queue_class_init (BognorQueueClass *klass)
{
    GObjectClass *o_class = (GObjectClass *) klass;

    o_class->dispose = bognor_queue_dispose;
    o_class->finalize = bognor_queue_finalize;
    o_class->set_property = bognor_queue_set_property;
    o_class->get_property = bognor_queue_get_property;

    g_type_class_add_private (klass, sizeof (BognorQueuePrivate));
    dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass),
                                     &dbus_glib_bognor_queue_object_info);

    g_object_class_install_property (o_class, PROP_NAME,
                                     g_param_spec_string ("name", "", "", "",
                                                          G_PARAM_READWRITE |
                                                          G_PARAM_CONSTRUCT_ONLY |
                                                          G_PARAM_STATIC_STRINGS));
    g_object_class_install_property (o_class, PROP_SOURCE,
                                     g_param_spec_object ("source-manager",
                                                          "", "",
                                                          BOGNOR_TYPE_SOURCE_MANAGER,
                                                          G_PARAM_CONSTRUCT_ONLY |
                                                          G_PARAM_STATIC_STRINGS |
                                                          G_PARAM_WRITABLE));

    signals[ADDED] = g_signal_new ("uri-added",
                                   G_TYPE_FROM_CLASS (klass),
                                   G_SIGNAL_RUN_FIRST |
                                   G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                   bognor_marshal_VOID__STRING_INT,
                                   G_TYPE_NONE, 2, G_TYPE_STRING,
                                   G_TYPE_INT);
    signals[REMOVED] = g_signal_new ("uri-removed",
                                     G_TYPE_FROM_CLASS (klass),
                                     G_SIGNAL_RUN_FIRST |
                                     G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                     bognor_marshal_VOID__STRING_INT,
                                     G_TYPE_NONE, 2, G_TYPE_STRING,
                                     G_TYPE_INT);
    signals[INDEX_CHANGED] = g_signal_new ("index-changed",
                                         G_TYPE_FROM_CLASS (klass),
                                         G_SIGNAL_RUN_FIRST |
                                         G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                         g_cclosure_marshal_VOID__INT,
                                         G_TYPE_NONE, 1, G_TYPE_INT);
    signals[POSITION_CHANGED] = g_signal_new ("position-changed",
                                              G_TYPE_FROM_CLASS (klass),
                                              G_SIGNAL_RUN_FIRST |
                                              G_SIGNAL_NO_RECURSE,
                                              0, NULL, NULL,
                                              g_cclosure_marshal_VOID__DOUBLE,
                                              G_TYPE_NONE, 1, G_TYPE_DOUBLE);
    signals[STATE_CHANGED] = g_signal_new ("state-changed",
                                             G_TYPE_FROM_CLASS (klass),
                                             G_SIGNAL_RUN_FIRST |
                                             G_SIGNAL_NO_RECURSE,
                                             0, NULL, NULL,
                                             g_cclosure_marshal_VOID__INT,
                                             G_TYPE_NONE, 1, G_TYPE_INT);
    signals[ITEM_CHANGED] = g_signal_new ("item-changed",
                                          G_TYPE_FROM_CLASS (klass),
                                          G_SIGNAL_RUN_FIRST |
                                          G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                          g_cclosure_marshal_VOID__OBJECT,
                                          G_TYPE_NONE, 1,
                                          BOGNOR_TYPE_QUEUE_ITEM);
}

static void
bognor_queue_init (BognorQueue *self)
{
    BognorQueuePrivate *priv = GET_PRIVATE (self);

    self->priv = priv;

    priv->play_queue = g_queue_new ();
    priv->current_position = NULL;
}

void
bognor_queue_notify_unknown_format (BognorQueue     *queue,
                                    BognorQueueItem *item)
{
    NotifyNotification *notification;
    const char *uri;
    char *basename, *unesc, *message;

    uri = bognor_queue_item_get_uri (item);
    basename = g_path_get_basename (uri);
    unesc = g_uri_unescape_string (basename, NULL);
    g_free (basename);

#ifdef ENABLE_HELIX
    message = g_strdup_printf (_("Sorry, we can't play %s, as we don't have the correct plugin."), unesc);
#else
    message = g_strdup_printf (_("Sorry, we can't play %s, as we don't have the correct plugin. You could try searching for ❝gstreamer codecs❞ on the web."), unesc);
#endif
    g_free (unesc);

    notification = notify_notification_new (_("We can't play this"),
                                            message, NULL, NULL);
    notify_notification_show (notification, NULL);
    g_object_unref (notification);

    g_free (message);
}

void
bognor_queue_emit_position_changed (BognorQueue *queue,
                                    double       position)
{
    g_signal_emit (queue, signals[POSITION_CHANGED], 0, position);
}


gboolean
bognor_queue_stop (BognorQueue *queue,
                   GError     **error)
{
    BognorQueueClass *klass = BOGNOR_QUEUE_GET_CLASS (queue);
    BognorQueuePrivate *priv = queue->priv;

    klass->set_playing (queue, FALSE);

    priv->state = BOGNOR_QUEUE_STATE_STOPPED;
    g_signal_emit (queue, signals[STATE_CHANGED], 0, priv->state);

    return TRUE;
}

gboolean
bognor_queue_play (BognorQueue *queue,
                   GError     **error)
{
    BognorQueueClass *klass = BOGNOR_QUEUE_GET_CLASS (queue);
    BognorQueuePrivate *priv = queue->priv;

    g_print ("Playing...\n");
    if (priv->current_position == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_EMPTY,
                              "The queue is empty");
        return FALSE;
    }

    klass->set_playing (queue, TRUE);

    priv->state = BOGNOR_QUEUE_STATE_PLAYING;
    g_signal_emit (queue, signals[STATE_CHANGED], 0, priv->state);

    return TRUE;
}

gboolean
bognor_queue_next (BognorQueue *queue,
                   GError     **error)
{
    BognorQueueClass *klass = BOGNOR_QUEUE_GET_CLASS (queue);
    BognorQueuePrivate *priv = queue->priv;

    if (priv->current_position == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_EMPTY,
                              "The queue is empty");
        return FALSE;
    }

    if (priv->current_position->next == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_END,
                              "The queue has finished");
        return FALSE;
    }

    priv->current_position = priv->current_position->next;
    priv->current_index++;

    klass->set_item (queue, (BognorQueueItem *) priv->current_position->data);
    g_signal_emit (queue, signals[INDEX_CHANGED], 0, priv->current_index);
    g_signal_emit (queue, signals[ITEM_CHANGED], 0,
                   priv->current_position->data);

    return TRUE;
}

gboolean
bognor_queue_previous (BognorQueue *queue,
                       GError     **error)
{
    BognorQueueClass *klass = BOGNOR_QUEUE_GET_CLASS (queue);
    BognorQueuePrivate *priv = queue->priv;

    if (priv->current_position == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_EMPTY,
                              "The queue is empty");
        return FALSE;
    }

    if (priv->current_position->prev == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_START,
                              "The queue is at the start");
        return FALSE;
    }

    priv->current_position = priv->current_position->prev;
    priv->current_index--;

    klass->set_item (queue, (BognorQueueItem *) priv->current_position->data);
    g_signal_emit (queue, signals[INDEX_CHANGED], 0, priv->current_index);
    g_signal_emit (queue, signals[ITEM_CHANGED], 0,
                   priv->current_position->data);

    return TRUE;
}

gboolean
bognor_queue_set_position (BognorQueue *queue,
                           double       position,
                           GError     **error)
{
    BognorQueueClass *klass = BOGNOR_QUEUE_GET_CLASS (queue);
    BognorQueuePrivate *priv = queue->priv;

    if (priv->current_position == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_EMPTY,
                              "The queue is empty");
        return FALSE;
    }

    klass->set_position (queue, position);
    return TRUE;
}

gboolean
bognor_queue_get_position (BognorQueue *queue,
                           double      *position,
                           GError     **error)
{
    BognorQueueClass *klass = BOGNOR_QUEUE_GET_CLASS (queue);
    BognorQueuePrivate *priv = queue->priv;

    if (priv->current_position == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_EMPTY,
                              "The queue is empty");
        *position = 0.0;
        return FALSE;
    }

    *position = klass->get_position (queue);

    return TRUE;
}

gboolean
bognor_queue_append_uris (BognorQueue *queue,
                          int          count,
                          const char **uris,
                          const char **mimetypes,
                          GError     **error)
{
    BognorQueuePrivate *priv = queue->priv;
    int i;

    for (i = 0; i < count; i++) {
        BklItem *bkl;
        BognorQueueItem *item;

        g_print ("Appending %s\n", uris[i]);
        bkl = bognor_source_manager_find_item (priv->source_manager, uris[i]);
        if (bkl == NULL) {
            item = bognor_queue_item_new (uris[i],
                                          mimetypes[i], NULL);
        } else {
            item = bognor_queue_item_new_from_bkl (bkl);
        }
        g_queue_push_tail (priv->play_queue, item);

        /* FIXME: Should we have an "add-start" and "add-finish" set
           of signals around this? */
        g_signal_emit (queue, signals[ADDED], 0, uris[i],
                       g_queue_get_length (priv->play_queue) - 1);
    }

    if (priv->current_position == NULL) {
        return bognor_queue_set_index (queue, 0, error);
    }

    return TRUE;
}

gboolean
bognor_queue_insert_uris (BognorQueue *queue,
                          int          position,
                          int          count,
                          const char **uris,
                          const char **mimetypes,
                          GError     **error)
{
    BognorQueuePrivate *priv = queue->priv;
    GList *nth;
    int i;

    nth = g_queue_peek_nth (priv->play_queue, position);
    if (nth == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_OUT_OF_RANGE,
                              "%d is out of range for queue with %d items",
                              position, g_queue_get_length (priv->play_queue));
        return FALSE;
    }

    /* FIXME: See issues in append_uris as well */
    for (i = 0; i < count; i++) {
        BklItem *bkl;
        BognorQueueItem *item;

        bkl = bognor_source_manager_find_item (priv->source_manager, uris[i]);
        if (bkl == NULL) {
            item = bognor_queue_item_new (uris[i],
                                          mimetypes[i], NULL);
        } else {
            item = bognor_queue_item_new_from_bkl (bkl);
        }
        g_queue_insert_before (priv->play_queue, nth, item);
        g_signal_emit (queue, signals[ADDED], 0, uris[i],
                       g_queue_get_length (priv->play_queue) - 1);
    }

    if (priv->current_position == NULL) {
        return bognor_queue_set_index (queue, 0, error);
    }

    return TRUE;
}

gboolean
bognor_queue_remove_range (BognorQueue *queue,
                           int          index,
                           int          count,
                           GError     **error)
{
    return TRUE;
}

gboolean
bognor_queue_set_index (BognorQueue *queue,
                        int          index,
                        GError     **error)
{
    BognorQueueClass *klass = BOGNOR_QUEUE_GET_CLASS (queue);
    BognorQueuePrivate *priv = queue->priv;

    g_print ("Setting index to %d\n", index);
    if (index == BOGNOR_QUEUE_INDEX_END) {
        priv->current_position = g_queue_peek_tail_link (priv->play_queue);
    } else {
        priv->current_position = g_queue_peek_nth_link (priv->play_queue, index);
    }

    if (priv->current_position == NULL && error != NULL) {
                *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_OUT_OF_RANGE,
                              "%d is out of range for queue with %d items",
                              index, g_queue_get_length (priv->play_queue));
        return FALSE;
    }

    klass->set_item (queue, priv->current_position->data);
    priv->current_index = index;

    if (index == BOGNOR_QUEUE_INDEX_END) {
        index = g_queue_get_length (priv->play_queue) - 1;
    }

    g_signal_emit (queue, signals[INDEX_CHANGED], 0, index);
    g_signal_emit (queue, signals[ITEM_CHANGED], 0,
                   priv->current_position->data);

    return TRUE;
}

gboolean
bognor_queue_get_index (BognorQueue *queue,
                        int         *index,
                        GError     **error)
{
    BognorQueuePrivate *priv = queue->priv;

    *index = priv->current_index;
    return TRUE;
}

gboolean
bognor_queue_get_index_uri (BognorQueue *queue,
                            int          index,
                            char       **uri,
                            char       **mimetype,
                            GError     **error)
{
    BognorQueuePrivate *priv = queue->priv;
    BognorQueueItem *item;

    item = g_queue_peek_nth (priv->play_queue, index);
    if (item == NULL && error != NULL) {
        *error = g_error_new (BOGNOR_QUEUE_ERROR,
                              BOGNOR_QUEUE_ERROR_OUT_OF_RANGE,
                              "%d is out of range for queue with %d items",
                              index, g_queue_get_length (priv->play_queue));
        return FALSE;
    }

    *uri = g_strdup (bognor_queue_item_get_uri (item));
    *mimetype = g_strdup (bognor_queue_item_get_mimetype (item));

    return TRUE;
}

gboolean
bognor_queue_list_uris (BognorQueue *queue,
                        char      ***uris,
                        GError     **error)
{
    BognorQueuePrivate *priv = queue->priv;
    char **uri_array;
    GList *u = NULL;
    int n_uris, i;

    n_uris = g_queue_get_length (priv->play_queue);
    uri_array = g_new (char *, n_uris + 1);

    for (u = priv->play_queue->head, i = 0; u; u = u->next, i++) {
        BognorQueueItem *item = u->data;

        uri_array[i] = g_strdup (bognor_queue_item_get_uri (item));
    }

    /* NULL terminate the array */
    uri_array[n_uris] = NULL;

    *uris = uri_array;

    return TRUE;
}

gboolean
bognor_queue_get_state (BognorQueue *queue,
                        int         *state,
                        GError     **error)
{
    BognorQueuePrivate *priv = queue->priv;

    *state = priv->state;
    return TRUE;
}

gboolean
bognor_queue_get_name (BognorQueue *queue,
                       char       **name,
                       GError     **error)
{
    BognorQueuePrivate *priv = queue->priv;

    *name = g_strdup (priv->name);
    return TRUE;
}

int
bognor_queue_get_count (BognorQueue *queue)
{
    BognorQueuePrivate *priv = queue->priv;

    return g_queue_get_length (priv->play_queue);
}
